@@ -12,13 +12,16 @@ const (
 	DefaultSocket = "/run/zsysd.sock"
 
 	// DefaultClientTimeout for client requests
-	DefaultClientTimeout = 30 * time.Second
+	DefaultClientTimeout = 120 * time.Second
 
 	// DefaultServerIdleTimeout is the default time without a request before the server exits
 	DefaultServerIdleTimeout = time.Minute
 
 	// DefaultPath is the default configuration path
 	DefaultPath = "/etc/zsys.conf"
+
+	// UserConfirmationNeeded is a dedicated type for GRPC error which signal that we need more info from user
+	UserConfirmationNeeded = "UserConfirmationNeeded"
 )
 
 var (
