@@ -3,6 +3,7 @@ package machines
 import (
 	"context"
 	"fmt"
+	"math"
 	"sort"
 	"strings"
 	"time"
@@ -26,18 +27,10 @@ type sortedReverseByTimeStates []*State
 
 func (s sortedReverseByTimeStates) Len() int { return len(s) }
 func (s sortedReverseByTimeStates) Less(i, j int) bool {
-	return s[i].LastUsed.After(*s[j].LastUsed)
+	return s[i].LastUsed.After(s[j].LastUsed)
 }
 func (s sortedReverseByTimeStates) Swap(i, j int) { s[i], s[j] = s[j], s[i] }
 
-type sortedReverseByTimeUserStates []UserState
-
-func (s sortedReverseByTimeUserStates) Len() int { return len(s) }
-func (s sortedReverseByTimeUserStates) Less(i, j int) bool {
-	return s[i].LastUsed.After(*s[j].LastUsed)
-}
-func (s sortedReverseByTimeUserStates) Swap(i, j int) { s[i], s[j] = s[j], s[i] }
-
 type keepStatus int
 
 const (
@@ -51,16 +44,12 @@ type stateWithKeep struct {
 	keep keepStatus
 }
 
-type userStateWithKeep struct {
-	*UserState
-	keep keepStatus
-}
-
 // GC starts garbage collection for system and users
+// If all is set manual snapshots are considered too
 func (ms *Machines) GC(ctx context.Context, all bool) error {
-	now := time.Now()
+	now := ms.time.Now()
 
-	buckets := computeBuckets(now, ms.conf.History)
+	buckets := computeBuckets(ctx, now, ms.conf.History)
 	keepLast := ms.conf.History.KeepLast
 
 	allDatasets := make([]*zfs.Dataset, 0, len(ms.allSystemDatasets)+len(ms.allPersistentDatasets)+len(ms.allUsersDatasets)+len(ms.unmanagedDatasets))
@@ -84,14 +73,20 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 	}
 
 	var statesToRemove []*State
+	keepDueToErrorOnDelete := make(map[string]bool)
 
 	// 1. System GC
-	log.Debug(ctx, i18n.G("System GC"))
-
+	var gcPassNum int
 	for {
+		gcPassNum++
+		log.Debugf(ctx, "GC System Pass #%d", gcPassNum)
 		statesChanges := false
 
 		for _, m := range ms.all {
+			if !m.isZsys() {
+				continue
+			}
+
 			var newestStateIndex int
 			var sortedStates sortedReverseByTimeStates
 			for _, s := range m.History {
@@ -104,7 +99,7 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 
 				// End of the array, nothing else to do.
 				if newestStateIndex >= len(sortedStates) {
-					log.Debug(ctx, i18n.G("No more system states left. Stopping analyzing buckets"))
+					log.Debugf(ctx, "No more system states left for pass #%d.", gcPassNum)
 					break
 				}
 
@@ -119,9 +114,11 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 
 				// Advance to first state matching this bucket.
 				var i int
-				for i = newestStateIndex; i < len(sortedStates) && sortedStates[i].LastUsed.After(bucket.start); i++ {
+				for i = newestStateIndex; i < len(sortedStates) && !sortedStates[i].LastUsed.Before(bucket.start); i++ {
 				}
+				// The condition became false on i (sortedStates[i].LastUsed is Before bucket.start), this the first one "oldest" matching this bucket is the previous one: i-1
 				oldestStateIndex := i - 1
+
 				log.Debugf(ctx, i18n.G("First state matching for this bucket: %s (%s)"), sortedStates[oldestStateIndex].ID, sortedStates[oldestStateIndex].LastUsed.Format(timeFormat))
 
 				// Don't touch anything for this bucket, skip all states in here and advance to next one.
@@ -146,10 +143,12 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 					} else if keep == keepUnknown && i < keepLast {
 						log.Debugf(ctx, i18n.G("Keeping snapshot %v as it's in the last %d snapshots"), s.ID, keepLast)
 						keep = keepYes
+					} else if keepDueToErrorOnDelete[s.ID] {
+						keep = keepYes
 					} else {
 						// We only collect systems because users will be untagged if they have any dependency
 					analyzeSystemDataset:
-						for _, ds := range s.SystemDatasets {
+						for _, ds := range s.Datasets {
 							for _, d := range ds {
 								// keep the whole state if any dataset is the origin of a clone of if it’s a clone with snapshots on it
 								if byOrigin[d.Name] != nil || snapshotsByDS[d.Name] != nil {
@@ -166,28 +165,24 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 						keep:  keep,
 					})
 				}
+				// next bucket start point
+				newestStateIndex = oldestStateIndex + 1
 
 				// Ensure we have the minimum amount of states on this bucket.
 				nStatesToRemove := len(states) - bucket.samples
 				if nStatesToRemove <= 0 {
 					log.Debugf(ctx, i18n.G("No exceeding states for this bucket (delta: %d). Moving on."), nStatesToRemove)
-					newestStateIndex = oldestStateIndex + 1
 					continue
 				}
 				log.Debugf(ctx, i18n.G("There are %d exceeding states to potentially remove"), nStatesToRemove)
 
-				// FIXME: easy path: Remove first states that we don't keep
-				for _, s := range states {
-					if s.keep == keepYes {
-						continue
-					}
-					if nStatesToRemove == 0 {
-						continue
-					}
+				statesToRemoveForBucket := selectStatesToRemove(ctx, bucket.samples, states)
 
+				for _, s := range statesToRemoveForBucket {
+					statesChanges = true
 					// We are removing that state: purge all datasets from our maps.
 					// We don’t deal with user datasets right now as we only untag them.
-					for _, ds := range s.SystemDatasets {
+					for _, ds := range s.Datasets {
 						for _, d := range ds {
 							if d.IsSnapshot {
 								n, _ := splitSnapshotName(d.Name)
@@ -199,12 +194,9 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 							}
 						}
 					}
-
-					statesToRemove = append(statesToRemove, s.State)
-					nStatesToRemove--
-					statesChanges = true
 				}
-				newestStateIndex = oldestStateIndex + 1
+
+				statesToRemove = append(statesToRemove, statesToRemoveForBucket...)
 			}
 		}
 
@@ -215,9 +207,10 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 
 		// Remove the given states.
 		for _, s := range statesToRemove {
-			log.Infof(ctx, i18n.G("Removing state: %s"), s.ID)
-			if err := s.Remove(ctx, ms.z); err != nil {
-				log.Errorf(ctx, i18n.G("Couldn't fully destroy state %s: %v"), s.ID, err)
+			log.Infof(ctx, i18n.G("Selecting state to remove: %s"), s.ID)
+			if err := s.remove(ctx, ms, ""); err != nil {
+				log.Errorf(ctx, i18n.G("Couldn't fully destroy state %s: %v\nPutting it in keep list."), s.ID, err)
+				keepDueToErrorOnDelete[s.ID] = true
 			}
 		}
 		statesToRemove = nil
@@ -227,23 +220,33 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 		log.Debug(ctx, i18n.G("System have changes, rerun system GC"))
 	}
 
-	// 2. GC user datasets
-	log.Debug(ctx, i18n.G("User GC"))
+	// 2. GC user datasets. Note that we will only collect user states that are independent of system states.
 	// TODO: this is a copy of above, but we keep any states associated with user states, we really need to merge State and UserStates
-	var UserStatesToRemove []*UserState
-
+	// FIXME: user states attached to multiple datasets are counted individually when removing user states, and so, we can think
+	// we keep more history than we will have in the end. We should only count them as a single one
+	statesToRemove = nil
+	// this is the map to preserve unassociated clones because they are filling the bucket policy
+	userDatasetsToKeep := make(map[string]bool)
+	keepDueToErrorOnDelete = make(map[string]bool)
+	gcPassNum = 0
 	for {
+		gcPassNum++
+		log.Debugf(ctx, "GC User Pass #%d", gcPassNum)
 		statesChanges := false
 
 		for _, m := range ms.all {
-			for _, us := range m.Users {
+			// FIXME: we count same user state multiple times if linked to multiple bootfs systems
+			for _, us := range m.AllUsersStates {
 				var newestStateIndex int
-				var sortedStates sortedReverseByTimeUserStates
+				var sortedStates sortedReverseByTimeStates
 
+			nextUserState:
 				for _, s := range us {
-					// exclude "current" user state
-					if _, exists := m.UserDatasets[s.ID]; exists {
-						continue
+					// exclude "current" user state fom history
+					for _, us := range m.State.Users {
+						if us == s {
+							continue nextUserState
+						}
 					}
 
 					sortedStates = append(sortedStates, s)
@@ -255,7 +258,7 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 
 					// End of the array, nothing else to do.
 					if newestStateIndex >= len(sortedStates) {
-						log.Debug(ctx, i18n.G("No more user states left. Stopping analyzing buckets"))
+						log.Debugf(ctx, "No more user states left for pass #%d.", gcPassNum)
 						break
 					}
 
@@ -283,24 +286,32 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 					}
 
 					// Collect all states for current bucket and mark those having constraints
-					states := make([]userStateWithKeep, 0, oldestStateIndex-newestStateIndex+1)
+					states := make([]stateWithKeep, 0, oldestStateIndex-newestStateIndex+1)
 					log.Debug(ctx, i18n.G("Collecting all states for current bucket"))
 					for i := newestStateIndex; i <= oldestStateIndex; i++ {
-						log.Debugf(ctx, i18n.G("Analyzing state %v: %v"), sortedStates[i].ID, sortedStates[i].LastUsed.Format(timeFormat))
 						s := sortedStates[i]
+						log.Debugf(ctx, i18n.G("Analyzing state %v: %v"), s.ID, s.LastUsed.Format(timeFormat))
 
 						keep := keepUnknown
-						// We can only collect snapshots here for user datasets, or they are unassociated clones that we will clean up later
-						if !s.isSnapshot() {
-							log.Debugf(ctx, i18n.G("Keeping %v as it's not a snapshot, and necessarily associated to a system state"), s.ID)
+						if !s.isSnapshot() && s.linkedToSystemState() {
+							log.Debugf(ctx, i18n.G("Keeping %v as it's not a snapshot and associated to a system state"), s.ID)
 							keep = keepYes
-						} else if keep == keepUnknown && !all && !strings.Contains(s.ID, "@"+automatedSnapshotPrefix) {
+						} else if !s.isSnapshot() {
+							for _, ds := range s.Datasets {
+								if ds[0].HasSnapshotInHierarchy() {
+									log.Debugf(ctx, i18n.G("Keeping %v as it has a snapshot in its child hierarchy"), s.ID)
+									keep = keepYes
+								}
+							}
+						} else if !all && s.isSnapshot() && !strings.Contains(s.ID, "@"+automatedSnapshotPrefix) {
 							log.Debugf(ctx, i18n.G("Keeping snapshot %v as it's not a zsys one"), s.ID)
 							keep = keepYes
-						} else if keep == keepUnknown && i < keepLast {
-							log.Debugf(ctx, i18n.G("Keeping snapshot %v as it's in the last %d snapshots"), s.ID, keepLast)
+						} else if i < keepLast {
+							log.Debugf(ctx, i18n.G("Keeping %v as it's in the last %d snapshots"), s.ID, keepLast)
 							keep = keepYes
-						} else if keep == keepUnknown {
+						} else if keepDueToErrorOnDelete[s.ID] {
+							keep = keepYes
+						} else {
 							_, snapshotName := splitSnapshotName(s.ID)
 							// Do we have a state associated with us?
 							for k, state := range m.History {
@@ -315,63 +326,63 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 									break
 								}
 							}
-						} else if keep == keepUnknown {
-							// check if any dataset has a automated or manual clone
-						analyzeUserDataset:
-							for _, d := range s.Datasets {
-								// We only treat snapshots as clones are necessarily associated with one system state or
-								// has already been destroyed and not associated.
-								// do we have clones of us?
-								if byOrigin[d.Name] != nil {
-									log.Debugf(ctx, i18n.G("Keeping snapshot %v as at least %s dataset has dependencies"), s.ID, d.Name)
-									keep = keepYes
-									break analyzeUserDataset
 
+							if keep == keepUnknown {
+								// check if any dataset has a automated or manual clone
+							analyzeUserDataset:
+								for _, ds := range s.Datasets {
+									for _, d := range ds {
+										// We only treat snapshots as clones are necessarily associated with one system state or
+										// has already been destroyed and not associated.
+										// do we have clones of us?
+										if byOrigin[d.Name] != nil {
+											log.Debugf(ctx, i18n.G("Keeping snapshot %v as at least %s dataset has dependencies"), s.ID, d.Name)
+											keep = keepYes
+											break analyzeUserDataset
+										}
+									}
 								}
 							}
 						}
 
-						states = append(states, userStateWithKeep{
-							UserState: &s,
-							keep:      keep,
+						states = append(states, stateWithKeep{
+							State: s,
+							keep:  keep,
 						})
+						for route := range s.Datasets {
+							userDatasetsToKeep[route] = true
+						}
 					}
+					// next bucket start point
+					newestStateIndex = oldestStateIndex + 1
 
 					// Ensure we have the minimum amount of states on this bucket.
 					nStatesToRemove := len(states) - bucket.samples
 					if nStatesToRemove <= 0 {
 						log.Debugf(ctx, i18n.G("No exceeding states for this bucket (delta: %d). Moving on."), nStatesToRemove)
-						newestStateIndex = oldestStateIndex + 1
 						continue
 					}
 					log.Debugf(ctx, i18n.G("There are %d exceeding states to potentially remove"), nStatesToRemove)
 
-					// FIXME: easy path: Remove first states that we don't keep
-					for _, s := range states {
-						if s.keep == keepYes {
-							continue
-						}
-						if nStatesToRemove == 0 {
-							continue
-						}
+					statesToRemoveForBucket := selectStatesToRemove(ctx, bucket.samples, states)
 
+					for _, s := range statesToRemoveForBucket {
+						statesChanges = true
 						// We are removing that state: purge all datasets from our maps.
-						for _, d := range s.Datasets {
-							if d.IsSnapshot {
-								n, _ := splitSnapshotName(d.Name)
-								snapshotsByDS[n] = removeFromSlice(snapshotsByDS[n], d.Name)
-							} else {
-								for orig := range byOrigin {
-									byOrigin[orig] = removeFromSlice(byOrigin[orig], d.Name)
+						for _, ds := range s.Datasets {
+							for _, d := range ds {
+								if d.IsSnapshot {
+									n, _ := splitSnapshotName(d.Name)
+									snapshotsByDS[n] = removeFromSlice(snapshotsByDS[n], d.Name)
+								} else {
+									for orig := range byOrigin {
+										byOrigin[orig] = removeFromSlice(byOrigin[orig], d.Name)
+									}
 								}
 							}
 						}
-
-						UserStatesToRemove = append(UserStatesToRemove, s.UserState)
-						nStatesToRemove--
-						statesChanges = true
 					}
-					newestStateIndex = oldestStateIndex + 1
+					statesToRemove = append(statesToRemove, statesToRemoveForBucket...)
 				}
 			}
 		}
@@ -381,41 +392,51 @@ func (ms *Machines) GC(ctx context.Context, all bool) error {
 		}
 
 		// Remove the given states.
-		nt := ms.z.NewNoTransaction(ctx)
-		for _, s := range UserStatesToRemove {
-			log.Infof(ctx, i18n.G("Removing state: %s"), s.ID)
-			if err := nt.Destroy(s.Datasets[0].Name); err != nil {
-				log.Errorf(ctx, i18n.G("Couldn't destroy user state %s: %v"), s, err)
+		for _, s := range statesToRemove {
+			log.Infof(ctx, i18n.G("Selecting state to remove: %s"), s.ID)
+			if err := s.remove(ctx, ms, ""); err != nil {
+				log.Errorf(ctx, i18n.G("Couldn't fully destroy user state %s: %v.\nPutting it in keep list."), s.ID, err)
+				keepDueToErrorOnDelete[s.ID] = true
+				continue
+			}
+			for route := range s.Datasets {
+				delete(userDatasetsToKeep, route)
 			}
 		}
 
-		UserStatesToRemove = nil
-		// TODO: only if changes
+		statesToRemove = nil
 		if err := ms.Refresh(ctx); err != nil {
 			return fmt.Errorf("Couldn't refresh machine list: %v", err)
 		}
 		log.Debug(ctx, i18n.G("Users states have changes, rerun user GC"))
 	}
 
-	// 3. Clean up user datasets with no tags. Take into account user datasets with a child not associated with anything but parent is
+	// 3. Clean up user user datasets with no tags. Take into account user datasets with a child not associated with anything but parent is
 	// (they, and all snapshots on them will end up in unmanaged datasets)
 	log.Debug(ctx, i18n.G("Unassociated user datasets GC"))
 	var alreadyDestroyedRoot []string
+	var hasChanges bool
 	nt := ms.z.NewNoTransaction(ctx)
 nextDataset:
-	for _, d := range ms.unmanagedDatasets {
-		if d.IsSnapshot || !isUserDataset(d.Name) {
+	for _, d := range ms.allUsersDatasets {
+		if d.IsSnapshot {
 			continue
 		}
 		if d.BootfsDatasets != "" {
 			continue
 		}
+		for rootDataset := range userDatasetsToKeep {
+			if d.Name == rootDataset || strings.HasPrefix(d.Name, rootDataset+"/") {
+				continue nextDataset
+			}
+		}
 		for _, n := range alreadyDestroyedRoot {
 			if strings.HasPrefix(d.Name, n+"/") {
 				continue nextDataset
 			}
 		}
 
+		hasChanges = true
 		// We destroy here all snapshots and leaf attached. Snapshots won’t be taken into account, however, we don’t want
 		// to try destroying leaves again, keep a list.
 		if err := nt.Destroy(d.Name); err != nil {
@@ -425,6 +446,12 @@ nextDataset:
 		alreadyDestroyedRoot = append(alreadyDestroyedRoot, d.Name)
 	}
 
+	if hasChanges {
+		if err := ms.Refresh(ctx); err != nil {
+			return fmt.Errorf("Couldn't refresh machine list: %v", err)
+		}
+	}
+
 	return nil
 }
 
@@ -453,7 +480,8 @@ func removeFromSlice(s []string, name string) (r []string) {
 
 // computeBuckets initializes the list of buckets in which the dataset will be sorted.
 // Buckets are defined from the main configuration file.
-func computeBuckets(now time.Time, rules config.HistoryRules) (buckets []bucket) {
+func computeBuckets(ctx context.Context, now time.Time, rules config.HistoryRules) (buckets []bucket) {
+	log.Debugf(ctx, "calculating buckets")
 	nowDay := time.Date(now.Year(), now.Month(), now.Day(), 0, 0, 0, 0, now.Location())
 	end := nowDay.Add(time.Duration(-rules.GCStartAfter * timeDay))
 
@@ -463,8 +491,10 @@ func computeBuckets(now time.Time, rules config.HistoryRules) (buckets []bucket)
 		end:     now,
 		samples: -1,
 	})
+	log.Debugf(ctx, "bucket keep all: start:%s, end:%s, samples:%d", buckets[len(buckets)-1].start, buckets[len(buckets)-1].end, buckets[len(buckets)-1].samples)
 
 	for _, rule := range rules.GCRules {
+		log.Debugf(ctx, "Rule %s, buckets: %d, length, %d", rule.Name, rule.Buckets, rule.BucketLength)
 		buckerDuration := time.Duration(timeDay * rule.BucketLength)
 
 		startPeriod := end.Add(-time.Duration(int(buckerDuration) * rule.Buckets))
@@ -475,6 +505,7 @@ func computeBuckets(now time.Time, rules config.HistoryRules) (buckets []bucket)
 				end:     d.Add(buckerDuration),
 				samples: rule.SamplesPerBucket,
 			})
+			log.Debugf(ctx, "  -  start:%s end:%s samples:%d", buckets[len(buckets)-1].start, buckets[len(buckets)-1].end, buckets[len(buckets)-1].samples)
 		}
 		end = startPeriod
 	}
@@ -485,6 +516,7 @@ func computeBuckets(now time.Time, rules config.HistoryRules) (buckets []bucket)
 		end:     end,
 		samples: 0,
 	})
+	log.Debugf(ctx, "bucket oldest: start:%s end:%s samples:%d", buckets[len(buckets)-1].start, buckets[len(buckets)-1].end, buckets[len(buckets)-1].samples)
 
 	return buckets
 }
@@ -513,3 +545,184 @@ func (b bucket) validate(oldState, newState []*stateWithKeep) bool {
 func (b bucket) String() string {
 	return fmt.Sprintf("start: %s end:%s samples: %d", b.start.Format(timeFormat), b.end.Format(timeFormat), b.samples)
 }
+
+// selectStatesToRemove selects the maximum number of states to keep to fill a bucket up to samples and spread them evenly over the width of the bucket.
+// When 2 solutions are equal the first match is kept
+func selectStatesToRemove(ctx context.Context, samples int, states []stateWithKeep) (statesToRemove []*State) {
+	log.Debug(ctx, "selecting list of states to remove")
+
+	if len(states) <= samples {
+		log.Debug(ctx, "bucket has enough capacity to keep all the states")
+		return statesToRemove
+	}
+
+	var sumKeep int64 // Start is oldest time and end is newest
+	var toPlace []stateWithKeep
+	var toKeep []stateWithKeep
+	var end, start float64
+
+	// States are supposed to be sorted by reverse time but we cannot assume it's true in this scope
+	for i, e := range states {
+		t := float64(e.LastUsed.Unix())
+		if i == 0 {
+			end = t
+			start = t
+		}
+		if t < start {
+			start = t
+		}
+		if t > end {
+			end = t
+		}
+	}
+	log.Debugf(ctx, "Interval: %.2f - %.2f", start, end)
+	for _, s := range states {
+		if s.keep == keepYes {
+			sumKeep += s.LastUsed.Unix()
+			toKeep = append(toKeep, s)
+			continue
+		}
+		toPlace = append(toPlace, s)
+	}
+
+	freeSlots := samples - len(toKeep)
+	if freeSlots <= 0 {
+		// The bucket is full, do not keep anything
+		log.Debug(ctx, "bucket is full, removing all non-keep states")
+		for _, s := range toPlace {
+			statesToRemove = append(statesToRemove, s.State)
+		}
+		return statesToRemove
+	}
+	cs := combinations(len(toPlace), freeSlots)
+
+	var bestCombination []int
+	var bestIndex int
+	minDistance := end - start
+	log.Debugf(ctx, "Existing n: %d, minDist: %.3f, barycenter: %.3f", len(toKeep), minDistance, start+minDistance/2)
+
+	var dbgMsg string
+	for i, c := range cs {
+		dbgMsg = fmt.Sprintf("    %d - ", i)
+
+		var sumToPlace int64
+		for i := 0; i < freeSlots; i++ {
+			sumToPlace += toPlace[c[i]].LastUsed.Unix()
+			dbgMsg += fmt.Sprintf("%d:%d ", c[i], toPlace[c[i]].LastUsed.Unix())
+		}
+
+		avg := float64(sumKeep+sumToPlace) / float64(samples)
+		curDistance := math.Abs(avg - (start + (end-start)/2))
+		if curDistance < minDistance {
+			minDistance = curDistance
+			bestCombination = c
+			bestIndex = i
+		}
+		log.Debugf(ctx, "%s%d %.3f %.3f", dbgMsg, sumToPlace, avg, curDistance)
+	}
+
+	dbgMsg = fmt.Sprintf("Best solution: dist=%.3f toPlace[%d] = [ ", minDistance, bestIndex)
+	for _, c := range bestCombination {
+		dbgMsg += fmt.Sprintf("%s:%d ", toPlace[c].ID, toPlace[c].LastUsed.Unix())
+	}
+	log.Debugf(ctx, "%s]", dbgMsg)
+
+	// Keep bestCombination and remove everything else not marked toKeep
+	for _, c := range bestCombination {
+		toPlace[c].keep = keepYes
+	}
+
+	for _, s := range toPlace {
+		if s.keep == keepYes {
+			continue
+		}
+		statesToRemove = append(statesToRemove, s.State)
+	}
+	return statesToRemove
+}
+
+// linkedToSystemState returns if a datasets is potentially linked to a system state.
+// Note that it doesn’t check if the system state is currently accessible.
+func (s *State) linkedToSystemState() bool {
+	for _, ds := range s.Datasets {
+		if ds[0].BootfsDatasets != "" {
+			return true
+		}
+	}
+	return false
+}
+
+const (
+	badNegInput = "combin: negative input"
+	badSetSize  = "combin: n < k"
+)
+
+// combinations generates all of the combinations of k elements from a
+// set of size n. The returned slice has length Binomial(n,k) and each inner slice
+// has length k.
+//
+// n and k must be non-negative with n >= k, otherwise Combinations will panic.
+//
+// This is copied from gonumonum
+func combinations(n, k int) [][]int {
+	combinations := binomial(n, k)
+	data := make([][]int, combinations)
+	if len(data) == 0 {
+		return data
+	}
+	data[0] = make([]int, k)
+	for i := range data[0] {
+		data[0][i] = i
+	}
+	for i := 1; i < combinations; i++ {
+		next := make([]int, k)
+		copy(next, data[i-1])
+		nextCombination(next, n, k)
+		data[i] = next
+	}
+	return data
+}
+
+// binomial returns the binomial coefficient of (n,k), also commonly referred to
+// as "n choose k".
+//
+// The binomial coefficient, C(n,k), is the number of unordered combinations of
+// k elements in a set that is n elements big, and is defined as
+//
+//  C(n,k) = n!/((n-k)!k!)
+//
+// n and k must be non-negative with n >= k, otherwise Binomial will panic.
+// No check is made for overflow.
+// This is copied from gonum
+func binomial(n, k int) int {
+	if n < 0 || k < 0 {
+		panic(badNegInput)
+	}
+	if n < k {
+		panic(badSetSize)
+	}
+	// (n,k) = (n, n-k)
+	if k > n/2 {
+		k = n - k
+	}
+	b := 1
+	for i := 1; i <= k; i++ {
+		b = (n - k + i) * b / i
+	}
+	return b
+}
+
+// nextCombination generates the combination after s, overwriting the input value.
+// This is copied from gonum
+func nextCombination(s []int, n, k int) {
+	for j := k - 1; j >= 0; j-- {
+		if s[j] == n+j-k {
+			continue
+		}
+		s[j]++
+		for l := j + 1; l < k; l++ {
+			s[l] = s[j] + l - j
+		}
+		break
+	}
+}
