@@ -4,9 +4,11 @@ import (
 	"context"
 	"errors"
 	"fmt"
+	"os"
 	"path/filepath"
 	"strconv"
 	"strings"
+	"syscall"
 	"time"
 
 	"github.com/ubuntu/zsys/internal/config"
@@ -45,13 +47,16 @@ func (ms *Machines) CreateUserData(ctx context.Context, user, homepath string) e
 
 	// Take same pool as existing userdatasets for current system
 	var userdatasetRoot string
-	for p := range ms.current.UserDatasets {
-		userdatasetRoot = getUserDatasetPath(p)
-		break
+selectUserDataset:
+	for _, us := range ms.current.Users {
+		for p := range us.Datasets {
+			userdatasetRoot = getUserDatasetRoot(p)
+			break selectUserDataset
+		}
 	}
 	// If there is none attached to the current system, try to take first existing userdataset detected pool
 	if userdatasetRoot == "" && len(ms.allUsersDatasets) > 0 {
-		userdatasetRoot = getUserDatasetPath(ms.allUsersDatasets[0].Name)
+		userdatasetRoot = getUserDatasetRoot(ms.allUsersDatasets[0].Name)
 	}
 
 	// If there is still none found, check if there is only USERDATA with no user under it as it won't shows up in machines
@@ -69,7 +74,7 @@ func (ms *Machines) CreateUserData(ctx context.Context, user, homepath string) e
 		p := ms.current.ID
 		i := strings.Index(p, "/")
 		if i != -1 {
-			p = p[0:i]
+			p = p[:i]
 		}
 		userdatasetRoot = filepath.Join(p, "USERDATA")
 
@@ -85,6 +90,10 @@ func (ms *Machines) CreateUserData(ctx context.Context, user, homepath string) e
 		cancel()
 		return err
 	}
+	// FIXME: mount the dataset here, we should have that in Create() but mitigate the impact for focal release
+	if err := syscall.Mount(userdataset, homepath, "zfs", 0, "zfsutil"); err != nil {
+		log.Warningf(ctx, i18n.G("Couldn't mount %s: %v"), homepath, err)
+	}
 
 	// Tag to associate with current system and lastUsed
 	if err := t.SetProperty(libzfs.BootfsDatasetsProp, ms.current.ID, userdataset, false); err != nil {
@@ -130,13 +139,13 @@ func (ms *Machines) ChangeHomeOnUserData(ctx context.Context, home, newHome stri
 	return ms.Refresh(ctx)
 }
 
-func getUserDatasetPath(path string) string {
+func getUserDatasetRoot(path string) string {
 	lpath := strings.ToLower(path)
 	i := strings.Index(lpath, userdatasetsContainerName)
 	if i == -1 {
 		return ""
 	}
-	return path[0 : i+len(userdatasetsContainerName)]
+	return path[:i+len(userdatasetsContainerName)]
 }
 
 // tryReuseUserDataSet tries to match an existing user dataset for the current machine.
@@ -145,37 +154,58 @@ func (ms *Machines) tryReuseUserDataSet(t *zfs.Transaction, user string, oldhome
 	log.Debugf(t.Context(), i18n.G("Trying to check if there is a user or home directory already attached to this machine"))
 
 	// If there this user or home already attached to this machine: retarget home
-	for _, ds := range ms.current.UserDatasets {
-		for _, d := range ds {
-			var match bool
-			// try handling user dataset
-			if user != "" {
-				// get user name from dataset
-				n := strings.Split(d.Name, "/")
-				userName := n[len(n)-1]
-				n = strings.Split(userName, "_")
-				userName = strings.Join(n[0:len(n)-1], "_")
+	for userName, us := range ms.current.Users {
+		for _, ds := range us.Datasets {
+			for _, d := range ds {
+				var match bool
+				// try handling user dataset
+				if user != "" {
+					// Home path is already attached to current system, but with a different user name. Fail
+					if d.Mountpoint == newhome && user != userName {
+						return false, fmt.Errorf(i18n.G("%q is already associated to %q, which is for a different user name (%q) than %q"), newhome, d.Name, userName, user)
+					}
+					if userName == user {
+						match = true
+					}
+				} else if oldhome != "" {
+					if d.Mountpoint == oldhome {
+						match = true
+					}
+				}
 
-				// Home path is already attached to current system, but with a different user name. Fail
-				if d.Mountpoint == newhome && user != userName {
-					return false, fmt.Errorf(i18n.G("%q is already associated to %q, which is for a different user name (%q) than %q"), newhome, d.Name, userName, user)
-				}
-				if userName == user {
-					match = true
-				}
-			} else if oldhome != "" {
-				if d.Mountpoint == oldhome {
-					match = true
-				}
-			}
+				// We'll reuse that dataset
+				if match {
+					log.Infof(t.Context(), i18n.G("Reusing %q as matching user name or old mountpoint"), d.Name)
+					var uid, gid int
+					if oldhome != "" {
+						if info, err := os.Stat(oldhome); err == nil {
+							if stat, ok := info.Sys().(*syscall.Stat_t); ok {
+								uid = int(stat.Uid)
+								gid = int(stat.Gid)
+							}
+						}
 
-			// We'll reuse that dataset
-			if match {
-				log.Infof(t.Context(), i18n.G("Reusing %q as matching user name or old mountpoint"), d.Name)
-				if err := t.SetProperty(libzfs.MountPointProp, newhome, d.Name, false); err != nil {
-					return false, fmt.Errorf(i18n.G("couldn't set new home %q to %q: ")+config.ErrorFormat, newhome, d.Name, err)
+					}
+					if err := t.SetProperty(libzfs.MountPointProp, newhome, d.Name, false); err != nil {
+						return false, fmt.Errorf(i18n.G("couldn't set new home %q to %q: ")+config.ErrorFormat, newhome, d.Name, err)
+					}
+					// Reset owner on newly created mountpoint
+					// FIXME: this should be in zfs itself when changing mount property and restore all properties on mountpoint itself
+					if uid != 0 {
+						if err := syscall.Unmount(newhome, 0); err == nil {
+							if err := os.Chown(newhome, uid, gid); err != nil {
+								log.Warningf(t.Context(), i18n.G("Couldn't restore permission on new home directory %s: %v"), newhome, err)
+							}
+							if err := syscall.Mount(d.Name, newhome, "zfs", 0, "zfsutil"); err != nil {
+								log.Warningf(t.Context(), i18n.G("Couldn't mount %s: %v"), newhome, err)
+							}
+						}
+					}
+					if err := os.Remove(oldhome); err != nil {
+						log.Warningf(t.Context(), i18n.G("couldn't cleanup %s directory: %v"), oldhome, err)
+					}
+					return true, nil
 				}
-				return true, nil
 			}
 		}
 	}
