@@ -62,8 +62,7 @@ func (d *Dataset) refreshProperties(ctx context.Context) error {
 	case "local":
 		sources.Mountpoint = "local"
 	case "default":
-		log.Warningf(ctx, i18n.G("MountPoint property for %q has an unexpected source: %q"), name, sourceMountPoint)
-		fallthrough
+		sources.Mountpoint = "inherited"
 	default:
 		if mountpoint != "" {
 			sources.Mountpoint = "inherited"
@@ -125,6 +124,15 @@ func (d *Dataset) refreshProperties(ctx context.Context) error {
 		if bootfsDatasets, srcBootfsDatasets, err = getUserPropertyFromSys(ctx, libzfs.BootfsDatasetsProp, d.dZFS); err != nil {
 			log.Warningf(ctx, i18n.G("can't read bootfsdataset property, ignoring: ")+config.ErrorFormat, err)
 		}
+		// TOREMOVE: transition from 19.10 installation without zsys to 20.04 after installing zsys
+		if strings.Contains(d.Name, "/USERDATA/") && bootfsDatasets == "" {
+			bootfsDatasets, srcBootfsDatasets, _ = getUserPropertyFromSys(ctx, "org.zsys:bootfs-datasets", d.dZFS)
+			if bootfsDatasets != "" {
+				if err := d.dZFS.SetUserProperty(libzfs.BootfsDatasetsProp, bootfsDatasets); err != nil {
+					log.Warningf(ctx, i18n.G("can't transition bootfsdataset property, ignoring: ")+config.ErrorFormat, err)
+				}
+			}
+		}
 	}
 	sources.BootfsDatasets = srcBootfsDatasets
 
@@ -158,7 +166,7 @@ func getUserPropertyFromSys(ctx context.Context, prop string, dZFS libzfs.DZFSIn
 	if p.Value == "-" && (p.Source == "-" || p.Source == "none") {
 		return "", "", nil
 	}
-	// The user property isn't set explicitely on the snapshot (inherited from non snapshot parent): ignore it.
+	// The user property isn't set explicitly on the snapshot (inherited from non snapshot parent): ignore it.
 	if dZFS.IsSnapshot() && p.Source != "local" {
 		return "", "", nil
 	}
@@ -186,6 +194,7 @@ func getUserPropertyFromSys(ctx context.Context, prop string, dZFS libzfs.DZFSIn
 }
 
 // newDatasetTree returns a Dataset and a populated tree of all its children
+// It returns a nil Dataset with a nil error for unsupported dataset type (DatasetTypeVolume or DatasetTypeBookmark)
 func newDatasetTree(ctx context.Context, dZFS libzfs.DZFSInterface, allDatasets *map[string]*Dataset) (*Dataset, error) {
 	// Skip non file system or snapshot datasets
 	if dZFS.Type() == libzfs.DatasetTypeVolume || dZFS.Type() == libzfs.DatasetTypeBookmark {
@@ -193,7 +202,7 @@ func newDatasetTree(ctx context.Context, dZFS libzfs.DZFSInterface, allDatasets
 	}
 
 	name := (*dZFS.Properties())[libzfs.DatasetPropName].Value
-	log.Debugf(ctx, i18n.G("New dataNew dataset found: %q"), name)
+	log.Debugf(ctx, i18n.G("New dataset found: %q"), name)
 	node := Dataset{
 		Name:       name,
 		IsSnapshot: dZFS.IsSnapshot(),
@@ -217,6 +226,10 @@ func newDatasetTree(ctx context.Context, dZFS libzfs.DZFSInterface, allDatasets
 		if err != nil {
 			return nil, fmt.Errorf("couldn't scan dataset: %v", err)
 		}
+		// Not a filesystem or snapshot dataset: skipping
+		if c == nil {
+			continue
+		}
 		children = append(children, c)
 	}
 	node.children = children
@@ -273,6 +286,19 @@ func (d Dataset) checkSnapshotHierarchyIntegrity(snapshotName string, snapshotOn
 	return nil
 }
 
+// HasSnapshotInHierarchy checks that the hierarchy of current dataset has a snapshot
+func (d Dataset) HasSnapshotInHierarchy() bool {
+	if d.IsSnapshot {
+		return false
+	}
+	for _, cd := range d.children {
+		if cd.IsSnapshot || cd.HasSnapshotInHierarchy() {
+			return true
+		}
+	}
+	return false
+}
+
 // checkNoClone checks that the hierarchy has no clone.
 func (d *Dataset) checkNoClone() error {
 	// TODO: this reopens the pool entirely, so can be a little bit slow. Could be reimplemented ourselves.
