@@ -22,6 +22,7 @@ type LibZFS struct {
 
 	errOnCreate       bool
 	errOnClone        bool
+	errOnDestroy      bool
 	errOnPromote      bool
 	errOnScan         bool
 	errOnSetProperty  bool
@@ -164,9 +165,15 @@ func (l *LibZFS) DatasetCreate(path string, dtype libzfs.DatasetType, props map[
 	if ok {
 		cm.Source = "local"
 	} else {
+		var v, s string
+		if dtype == libzfs.DatasetTypeFilesystem {
+			v = "on"
+			s = "default"
+		}
+
 		cm = libzfs.Property{
-			Value:  "on",
-			Source: "default",
+			Value:  v,
+			Source: s,
 		}
 	}
 	props[libzfs.DatasetPropCanmount] = cm
@@ -184,7 +191,7 @@ func (l *LibZFS) DatasetCreate(path string, dtype libzfs.DatasetType, props map[
 		// ZFS Properties
 		pprops := parent.Dataset.Properties
 		for k, pp := range pprops {
-			// Overriden local property
+			// Overridden local property
 			if _, ok := props[k]; ok {
 				p := props[k]
 				if p.Source == "" {
@@ -218,6 +225,13 @@ func (l *LibZFS) DatasetCreate(path string, dtype libzfs.DatasetType, props map[
 				userProperties[k] = p
 			}
 		}
+	} else {
+		if _, ok := props[libzfs.DatasetPropMountpoint]; !ok {
+			props[libzfs.DatasetPropMountpoint] = libzfs.Property{
+				Value:  "/" + path,
+				Source: "default",
+			}
+		}
 	}
 
 	d := dZFS{
@@ -253,14 +267,14 @@ func (l *LibZFS) DatasetCreate(path string, dtype libzfs.DatasetType, props map[
 }
 
 // DatasetSnapshot creates a snapshot
-func (l *LibZFS) DatasetSnapshot(path string, recur bool, props map[libzfs.Prop]libzfs.Property) (libzfs.DZFSInterface, error) {
+func (l *LibZFS) DatasetSnapshot(path string, recur bool, props map[libzfs.Prop]libzfs.Property, userProps map[string]string) (libzfs.DZFSInterface, error) {
 	if len(strings.Split(path, "@")) != 2 || strings.Split(path, "@")[1] == "" {
 		return nil, fmt.Errorf("%q is not a valid snapshot name", path)
 	}
-	return l.createSnapshot(path, recur, props)
+	return l.createSnapshot(path, recur, props, userProps)
 }
 
-func (l *LibZFS) createSnapshot(path string, recur bool, props map[libzfs.Prop]libzfs.Property) (libzfs.DZFSInterface, error) {
+func (l *LibZFS) createSnapshot(path string, recur bool, props map[libzfs.Prop]libzfs.Property, userProps map[string]string) (libzfs.DZFSInterface, error) {
 	if l.forceLastUsedTime {
 		props[libzfs.DatasetPropCreation] = libzfs.Property{Value: currentMagicTime}
 	}
@@ -271,6 +285,11 @@ func (l *LibZFS) createSnapshot(path string, recur bool, props map[libzfs.Prop]l
 	}
 
 	d := dinterface.(*dZFS)
+	for k, v := range userProps {
+		if err := d.SetUserProperty(k, v); err != nil {
+			return nil, err
+		}
+	}
 
 	if !recur {
 		return d, nil
@@ -282,7 +301,7 @@ func (l *LibZFS) createSnapshot(path string, recur bool, props map[libzfs.Prop]l
 		}
 
 		childPath := c.Dataset.Properties[libzfs.DatasetPropName].Value + "@" + snapshotName
-		_, err := l.createSnapshot(childPath, recur, props)
+		_, err := l.createSnapshot(childPath, recur, props, userProps)
 		if err != nil {
 			return nil, err
 		}
@@ -327,6 +346,11 @@ func (l *LibZFS) ErrOnCreate(shouldErr bool) {
 	l.errOnCreate = shouldErr
 }
 
+// ErrOnDestroy forces a failure of the mock on destroy operation
+func (l *LibZFS) ErrOnDestroy(shouldErr bool) {
+	l.errOnDestroy = shouldErr
+}
+
 // ForceLastUsedTime ensures that any LastUsed property is set to the magic time for reproducibility
 func (l *LibZFS) ForceLastUsedTime(force bool) {
 	l.forceLastUsedTime = force
@@ -484,10 +508,26 @@ func (d *dZFS) setPropertyWithSource(p libzfs.Prop, value, source string) error
 
 func (d *dZFS) Destroy(Defer bool) (err error) {
 	d.assertDatasetOpened()
+	n := d.Dataset.Properties[libzfs.DatasetPropName].Value
+
+	if d.libZFSMock.errOnDestroy {
+		return errors.New("Error on Destroy requested")
+	}
 
 	d.libZFSMock.mu.Lock()
 	defer d.libZFSMock.mu.Unlock()
-	delete(d.libZFSMock.datasets, d.Dataset.Properties[libzfs.DatasetPropName].Value)
+	for name, dataset := range d.libZFSMock.datasets {
+		if n == name {
+			continue
+		}
+		if strings.HasPrefix(name, n+"/") || strings.HasPrefix(name, n+"@") {
+			return fmt.Errorf("can't remove %s: it has at least one child: %s", n, name)
+		}
+		if dataset.Dataset.Properties[libzfs.DatasetPropOrigin].Value == n {
+			return fmt.Errorf("can't remove %s: it has at least one clone: %s", n, name)
+		}
+	}
+	delete(d.libZFSMock.datasets, n)
 	return nil
 }
 
@@ -556,7 +596,7 @@ func (d *dZFS) Promote() (err error) {
 		oldDatasetName := snap.Dataset.Properties[libzfs.DatasetPropName].Value
 		newName := datasetName + "@" + strings.Split(oldDatasetName, "@")[1]
 
-		// Pass a copy of properties to not alter - soon deleted - old snapshot on previoulsy promoted dataset
+		// Pass a copy of properties to not alter - soon deleted - old snapshot on previously promoted dataset
 		newDProps := make(map[libzfs.Prop]libzfs.Property)
 		for k, v := range snap.Dataset.Properties {
 			newDProps[k] = v
