@@ -123,7 +123,7 @@ func (z *Zfs) Refresh(ctx context.Context) error {
 }
 
 // Datasets returns all datasets on the system, where parent will always be before children.
-func (z Zfs) Datasets() []Dataset {
+func (z Zfs) Datasets() []*Dataset {
 	ds := make(chan *Dataset)
 
 	var collectChildren func(d *Dataset)
@@ -140,9 +140,9 @@ func (z Zfs) Datasets() []Dataset {
 	}
 	go collectChildren(z.root)
 
-	r := make([]Dataset, 0, len(z.allDatasets))
+	r := make([]*Dataset, 0, len(z.allDatasets))
 	for d := range ds {
-		r = append(r, *d)
+		r = append(r, d)
 	}
 	return r
 }
@@ -226,6 +226,11 @@ func (t *Transaction) Context() context.Context {
 	return t.ctx
 }
 
+// Context returns the current context of the transaction
+func (nt *NoTransaction) Context() context.Context {
+	return nt.ctx
+}
+
 // findDatasetByName returns given dataset from path, handling special case of dataset
 // not found in hashmap and virtual root dataset.
 func (z *Zfs) findDatasetByName(path string) (*Dataset, error) {
@@ -291,12 +296,12 @@ func (t *Transaction) newNestedTransaction() *nestedTransaction {
 	}
 }
 
-// Done either commit a nested transaction or cancel it if an error occured
+// Done either commit a nested transaction or cancel it if an error occurred
 func (t *nestedTransaction) Done(err *error) {
 	defer t.Transaction.Done()
 	if *err != nil {
 		// revert all in progress transactions
-		log.Debugf(t.ctx, i18n.G("ZFS: an error occured: %v"), *err)
+		log.Debugf(t.ctx, i18n.G("ZFS: an error occurred: %v"), *err)
 		log.Debugf(t.ctx, i18n.G("ZFS: Cancelling nested transaction"))
 		t.cancel()
 		return
@@ -377,25 +382,6 @@ func (t *nestedTransaction) snapshotRecursive(parent *Dataset, snapName string,
 
 	props := make(map[libzfs.Prop]libzfs.Property)
 
-	dZFS, err := t.Zfs.libzfs.DatasetSnapshot(parent.Name+"@"+snapName, false, props)
-	if err != nil {
-		return fmt.Errorf(i18n.G("couldn't create snapshot %q: %v"), parent.Name+"@"+snapName, err)
-	}
-
-	d := Dataset{
-		Name:       parent.Name + "@" + snapName,
-		IsSnapshot: true,
-		dZFS:       dZFS,
-	}
-	t.registerRevert(func() error {
-		nt := t.Zfs.NewNoTransaction(t.ctx)
-		if err := nt.destroyOne(&d); err != nil {
-			return fmt.Errorf(i18n.G("couldn't destroy %q for cleanup: %v"), d.Name, err)
-		}
-		return nil
-	})
-
-	// Set user properties that we couldn't set before creating the snapshot dataset.
 	// We don't set LastUsed here as Creation time will be used.
 	userPropertiesToSet := map[string]string{
 		libzfs.SnapshotMountpointProp: srcProps.Mountpoint + ":" + srcProps.sources.Mountpoint,
@@ -413,16 +399,24 @@ func (t *nestedTransaction) snapshotRecursive(parent *Dataset, snapName string,
 		userPropertiesToSet[libzfs.LastBootedKernelProp] = srcProps.LastBootedKernel + ":" + srcProps.sources.LastBootedKernel
 	}
 
-	for prop, value := range userPropertiesToSet {
-		// Only set values that are not empty on parent
-		if strings.HasPrefix(value, ":") {
-			continue
-		}
-		if err := dZFS.SetUserProperty(prop, value); err != nil {
-			return fmt.Errorf(i18n.G("couldn't set user property %q to %q for %v: ")+config.ErrorFormat, prop, value, d.Name, err)
-		}
+	dZFS, err := t.Zfs.libzfs.DatasetSnapshot(parent.Name+"@"+snapName, false, props, userPropertiesToSet)
+	if err != nil {
+		return fmt.Errorf(i18n.G("couldn't create snapshot %q: %v"), parent.Name+"@"+snapName, err)
 	}
 
+	d := Dataset{
+		Name:       parent.Name + "@" + snapName,
+		IsSnapshot: true,
+		dZFS:       dZFS,
+	}
+	t.registerRevert(func() error {
+		nt := t.Zfs.NewNoTransaction(t.ctx)
+		if err := nt.destroyOne(&d); err != nil {
+			return fmt.Errorf(i18n.G("couldn't destroy %q for cleanup: %v"), d.Name, err)
+		}
+		return nil
+	})
+
 	if err := d.refreshProperties(t.ctx); err != nil {
 		log.Warningf(t.ctx, i18n.G("couldn't fetch property of newly created snapshot: %v"), err)
 	}
@@ -552,7 +546,7 @@ func (t *nestedTransaction) cloneDataset(d Dataset, target string, ignoreErrorOn
 	}
 
 	// We want CanMount on -> noauto (don't mount new cloned dataset on top of parent)
-	// Otherwise, only set explicitely local property
+	// Otherwise, only set explicitly local property
 	if d.sources.CanMount == "local" || d.CanMount == "on" {
 		value := d.CanMount
 		if value == "on" {
@@ -709,6 +703,7 @@ func (t *nestedTransaction) promoteRecursive(d *Dataset) error {
 }
 
 // Destroy recursively all children, including dataset named "name".
+// If the dataset is a filesystem dataset, only remove it and children if there is no snapshots in the descendants.
 // If the dataset is a snapshot, navigate through the hierarchy to delete all dataset with the same snapshot name.
 // Note that destruction can't be rollbacked as filesystem content can't be recreated, so we don't accept them
 // in a transactional Zfs element.
@@ -732,6 +727,10 @@ func (nt *NoTransaction) Destroy(name string) error {
 			return fmt.Errorf(i18n.G("cannot find parent for %q: %v"), d.Name, err)
 		}
 	}
+
+	if d.HasSnapshotInHierarchy() {
+		return fmt.Errorf(i18n.G("couldn't destroy %q: it's a filesystem dataset which has snapshots"), d.Name)
+	}
 	if err := nt.destroyRecursive(target, snapName); err != nil {
 		return fmt.Errorf(i18n.G("couldn't destroy %q and its children: %v"), name, err)
 	}
@@ -741,7 +740,12 @@ func (nt *NoTransaction) Destroy(name string) error {
 
 // destroyRecursive destroys and unreference dataset objects, starting with children.
 func (nt *NoTransaction) destroyRecursive(d *Dataset, snapName string) error {
-	log.Debugf(nt.ctx, i18n.G("ZFS: trying to destroy recursively %q @ %q"), d.Name, snapName)
+	// Only try to find snapshot to destroy on filesystem datasets (snapshots can’t have snapshot as child)
+	if d.IsSnapshot && snapName != "" {
+		return nil
+	}
+
+	log.Debugf(nt.ctx, i18n.G("ZFS: trying to destroy recursively %s@%s"), d.Name, snapName)
 	copied := make([]*Dataset, len(d.children))
 	copy(copied, d.children)
 	for _, dc := range copied {
@@ -823,3 +827,63 @@ func (t *Transaction) SetProperty(name, value, datasetName string, force bool) e
 
 	return nil
 }
+
+// Dependencies returns the list of dataset dependencies in reverse order (deepest first)
+// A dataset has dependencies if:
+//   - it has a subdataset (has child)
+//   - it has a snapshot (so has child as well)
+//   - there is a clone depending on it (clone depending on snapshot on this dataset,
+//     so meaning that this dataset has a snapshot, so has child as well)
+func (nt *NoTransaction) Dependencies(d Dataset) []*Dataset {
+	ctx := nt.ctx
+
+	var deps []*Dataset
+	base, snapshot := splitSnapshotName(d.Name)
+	log.Debugf(ctx, "calculating dependencies for dataset %s", d.Name)
+
+	for k, dataset := range nt.Zfs.allDatasets {
+		var isDep bool
+		if !d.IsSnapshot {
+			// snapshot on filesystem dataset
+			if strings.HasPrefix(k, d.Name+"@") {
+				log.Debugf(ctx, "found dependency: snapshot %s", k)
+				isDep = true
+			}
+			// direct child
+			if strings.HasPrefix(k, d.Name+"/") && !strings.Contains(strings.TrimPrefix(k, d.Name+"/"), "/") {
+				log.Debugf(ctx, "found dependency: dkirect child %s", k)
+				isDep = true
+			}
+		} else {
+			if dataset.Origin == d.Name {
+				log.Debugf(ctx, "found dependency: origin %s", dataset.Name)
+				isDep = true
+			}
+			// Consider snapshot child: if d is a snapshot, look if subdataset which has the same base parent
+			// finish with the same snapshot name and consider them linked.
+			if dataset.IsSnapshot {
+				baseDataset, snapshotDataset := splitSnapshotName(dataset.Name)
+				if snapshotDataset == snapshot && strings.HasPrefix(baseDataset, base+"/") {
+					log.Debugf(ctx, "found dependency: snapshot child %s", dataset.Name)
+					isDep = true
+				}
+			}
+		}
+
+		if isDep {
+			deps = append(deps, nt.Dependencies(*dataset)...)
+			deps = append(deps, dataset)
+		}
+	}
+
+	// Deduplicate dependencies, keeping first which will has its inverse deps just after
+	keys := make(map[*Dataset]bool)
+	var uniqDeps []*Dataset
+	for _, entry := range deps {
+		if _, value := keys[entry]; !value {
+			keys[entry] = true
+			uniqDeps = append(uniqDeps, entry)
+		}
+	}
+	return uniqDeps
+}
